// SPDX-License-Identifier: GPL-2.0-only

/*
 *  ZeroMAC directory list utility
 *
 *  Author:
 *      Peter Gabor Gyulay <peter.gabor.gyulay@zeromac.uk>
 *
 *  Copyright (C) 2015-2020 Peter Gabor Gyulay <peter.gabor.gyulay@zeromac.uk>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License version 2,
 *      as published by the Free Software Foundation.
 */
           
           
#define LIB_ZMAC_LABEL_INLINE // prefer some functions inline

#include "../lib/libzmac_label.c"

char *filterfilename=NULL;

int zmac_ls_filterproc (const struct dirent *ddd)
{
	if (!filterfilename || !*filterfilename) return 1;
	register int c;
	for (c=0;ddd->d_name[c] && filterfilename[c] && (ddd->d_name[c]==filterfilename[c]);c++);
	return (!ddd->d_name[c] && !filterfilename[c]);
}

void zmac_ls_printxattr(char *fn, mode_t m)
{
	char label[MAX_STORAGE_LABEL_LENGTH+1], *exec=NULL;
	register int len, c, p;
	
	if (zmac_getlabel(fn, label) && *label)	
	{
		len=strlen(label);
		for (p=0;label[p];p++) if (label[p]==':') break;
		if (p<len-1) exec=label+p+1;
		c=10-p; if (c<0) c=0;
		while (c--) printf(" ");
		printf("%s", label);
		c=11-(len-p); if (c<0) c=0;
		while (c--) printf(" ");
		memset(label, 0, len);
	}
	else printf("                     ");								// 21 spaces

	if (S_ISDIR(m) && exec) printf(" S "); else printf("   ");
}

const char *zmac_ls_perms(char *buff, mode_t m)  
{  
	register char *b=buff;
	*b++=S_ISREG(m)?'-':S_ISSOCK(m)?'s':S_ISLNK(m)?'l':S_ISDIR(m)?'d':S_ISBLK(m)?'b':S_ISCHR(m)?'c':S_ISFIFO(m)?'|':'?';
	*b++=m&S_IRUSR?'r':'-';
	*b++=m&S_IWUSR?'w':'-';
	*b++=m&S_ISUID?'S':m&S_IXUSR?'x':'-';
	*b++=m&S_IRGRP?'r':'-';
	*b++=m&S_IWGRP?'w':'-';
	*b++=m&S_ISGID?'S':m&S_IXGRP?'x':'-';
	*b++=m&S_IROTH?'r':'-';
	*b++=m&S_IWOTH?'w':'-';
	*b++=m&S_IXOTH?'x':'-'; 
	*b=0;
	return buff;  
}  

  
int zmac_ls(char *pathname, char *filename)
{
	filterfilename=filename;
	int count=0, i, l;
	struct direct **files;  
	struct stat statbuf;  
	struct passwd pwent, *pwentp;  
	struct group grp, *grpp;  
	struct tm time;  
	char linknamebuff[MAXPATHLEN+1];
	char datestring[256], buf[65535], pbuff[12];  
	count=scandir(".", &files, zmac_ls_filterproc, alphasort);  
	if(count>0)  
	{  
		for (i=0; i<count; ++i)  
		{  
			if (lstat(files[i]->d_name, &statbuf) == 0)  
			{  
				zmac_ls_printxattr(files[i]->d_name, statbuf.st_mode);			// xattr
				printf("%10.10s", zmac_ls_perms(pbuff, statbuf.st_mode));		// file permission
				printf(" %3d", (int)statbuf.st_nlink);					// number of hardlinks
				if (!getpwuid_r(statbuf.st_uid, &pwent, buf, sizeof(buf), &pwentp)) printf(" %s", pwent.pw_name); else printf(" %d", statbuf.st_uid);	// user
				if (!getgrgid_r(statbuf.st_gid, &grp, buf, sizeof(buf), &grpp)) printf(" %s", grp.gr_name); else printf(" (%d)", statbuf.st_gid);	// group
				printf(" %8d", (int)statbuf.st_size);					// filesize
				localtime_r(&statbuf.st_mtime, &time);	
				strftime(datestring, sizeof(datestring), "%F %T", &time);		// time
				printf(" %s %s/%s", datestring, (strlen(pathname)==1)&& *pathname=='/'?"":pathname, files[i]->d_name);	// path, filename 
				if (S_ISLNK(statbuf.st_mode)) 
				{

					if ((l=readlink(files[i]->d_name, linknamebuff, MAXPATHLEN))>0)
					{
						linknamebuff[l]=0;
						printf(" -> %s", linknamebuff);
					}
				}
				printf("\n");
			}  
			free (files[i]);  
		}  
		free(files);  
	}  
	return count;
}

// new gcc renders libc strncpy, strncat useless by warnings of normal operation
void strncpy2(char *dst, char *src, size_t size)
{
	if (!size--) return;
	while (size--) if (!(*dst++=*src++)) return;
	*dst=0;
	return;
}

// new gcc renders libc strncpy, strncat useless by warnings of normal operation
void strncat2(char *dst, char *src, size_t size)
{
	size_t pos=0;
	while (dst[pos]) pos++;
	if (size<=pos+1) return; 
	size-=pos+1;
	dst+=pos;
	while (size--) if (!(*dst++=*src++)) return;
	*dst=0;
	return;
}


int main(int argc, char **argv, char **env)
{  
	char filename[MAXPATHLEN]={0}, cwd[MAXPATHLEN], pathname[MAXPATHLEN]={0}, fullpath[MAXPATHLEN]={0}, prevpath[MAXPATHLEN]={0}, lspath[MAXPATHLEN]={0};  
	register int c, d, e;
	if (!getcwd(cwd, MAXPATHLEN)) {printf("Cannot get current working directory!\n"); exit(1);}
	if (argc<2) {*filename=0; zmac_ls(cwd, NULL);}
	for (c=1;c<argc;c++)
	{
		*pathname=0; *filename=0;
		strncpy2(pathname, argv[c], MAXPATHLEN);
		for (e=-1, d=0;pathname[d];d++) if (pathname[d]=='/') e=d;				// where is the last forward slash
		if (e>=0)										// if we have path
		{	
			if (!e) {strncpy2(filename, pathname+1, MAXPATHLEN);pathname[1]=0;}		// if the full path is a slash, this is the root
			else {pathname[e]=0; strncpy2(filename, pathname+e+1, MAXPATHLEN);}		// leave path in pathname, copy filename into place
		}
		else {strncpy2(filename, pathname, MAXPATHLEN); *pathname=0;}				// if there is no path, the full string is a filename
		if (*pathname=='/') strncpy2(fullpath, pathname, MAXPATHLEN);				// if we get absolute path
		else  { strncpy2(fullpath, cwd, MAXPATHLEN); if (strlen(cwd)!=1)  strncat2(fullpath, "/", MAXPATHLEN);  strncat2(fullpath, pathname, MAXPATHLEN);}  // if we get relative path

		if (!*prevpath || strncmp(prevpath, fullpath, MAXPATHLEN))				// if no prev. path, or it is not the same as before, 
		{
			if (chdir(fullpath)) {printf("invalid path: %s\n", fullpath); continue;}
			if (!getcwd(lspath, MAXPATHLEN)) {printf("Cannot get directory name! (%s)\n", fullpath); continue;}
			strncpy2(prevpath, fullpath, MAXPATHLEN);
		}
		zmac_ls(lspath, filename);
	}
	return 0;
} 


